/*
 * bme280.c
 *
 *  Created on: May 05, 2025
 *      Author: Alex - www.aeq-web.com
 */

#include "bme280.h"
#include "i2c.h"
#include "main.h"
#include "hw_gpio.h"

static uint8_t bme280_address = 0x76;

// Calibration values
static uint16_t dig_T1;
static int16_t dig_T2, dig_T3;
static uint16_t dig_P1;
static int16_t dig_P2, dig_P3, dig_P4, dig_P5, dig_P6, dig_P7, dig_P8, dig_P9;
static uint8_t dig_H1;
static int16_t dig_H2;
static uint8_t dig_H3;
static int16_t dig_H4, dig_H5;
static int8_t dig_H6;

void bme280_init(uint8_t addr) {

	GPIO_InitTypeDef pinConfig = { 0 };
	pinConfig.Mode = GPIO_MODE_OUTPUT_PP;
	pinConfig.Pull = GPIO_NOPULL;
	pinConfig.Speed = GPIO_SPEED_FREQ_VERY_HIGH;

	HW_GPIO_Init( GPIOC, GPIO_PIN_13, &pinConfig);
	HW_GPIO_Write( GPIOC, GPIO_PIN_13, 1);

	HAL_Delay(2);

	bme280_address = addr;
	i2c_mod2_init();
	HAL_Delay(1);

}

void bme280_deinit(void) {
	i2c_mod2_deinit();
	HW_GPIO_Deinit( GPIOC, GPIO_PIN_13, 0);
}

bool bme280_connected(void) {

	uint8_t chip_id = 0;
	i2c_mod2_read_register(bme280_address << 1, 0xD0, 1, &chip_id, 1);

	if (chip_id == 0x60) {
		return true;
	}

	return false;
}

bool bme280_read_calibration(void) {
	uint8_t buf[26];

	if (i2c_mod2_read_register(bme280_address << 1, 0x88, 1, buf, 26) != HAL_OK) {
		return false;
	}

	dig_T1 = (uint16_t) (buf[1] << 8 | buf[0]);
	dig_T2 = (int16_t) (buf[3] << 8 | buf[2]);
	dig_T3 = (int16_t) (buf[5] << 8 | buf[4]);
	dig_P1 = (uint16_t) (buf[7] << 8 | buf[6]);
	dig_P2 = (int16_t) (buf[9] << 8 | buf[8]);
	dig_P3 = (int16_t) (buf[11] << 8 | buf[10]);
	dig_P4 = (int16_t) (buf[13] << 8 | buf[12]);
	dig_P5 = (int16_t) (buf[15] << 8 | buf[14]);
	dig_P6 = (int16_t) (buf[17] << 8 | buf[16]);
	dig_P7 = (int16_t) (buf[19] << 8 | buf[18]);
	dig_P8 = (int16_t) (buf[21] << 8 | buf[20]);
	dig_P9 = (int16_t) (buf[23] << 8 | buf[22]);

	uint8_t hbuf[7];

	if (i2c_mod2_read_register(bme280_address << 1, 0xA1, 1, hbuf, 1) != HAL_OK) {
		return false;
	}

	dig_H1 = hbuf[0];

	if (i2c_mod2_read_register(bme280_address << 1, 0xE1, 1, hbuf, 7) != HAL_OK) {
		return false;
	}

	dig_H2 = (int16_t) (hbuf[1] << 8 | hbuf[0]);
	dig_H3 = hbuf[2];
	dig_H4 = (int16_t) ((hbuf[3] << 4) | (hbuf[4] & 0x0F));
	dig_H5 = (int16_t) ((hbuf[5] << 4) | (hbuf[4] >> 4));
	dig_H6 = (int8_t) hbuf[6];

	return true;
}

float bme280_read_temperature(void) {
	uint8_t ctrl_meas = 0b00100101;
	i2c_mod2_write_register(bme280_address << 1, 0xF4, 1, &ctrl_meas, 1);
	HAL_Delay(10);

	uint8_t buf[3];
	i2c_mod2_read_register(bme280_address << 1, 0xFA, 1, buf, 3);

	int32_t adc_T = ((int32_t) buf[0] << 12) | ((int32_t) buf[1] << 4) | (buf[2] >> 4);

	int32_t var1 = ((((adc_T >> 3) - ((int32_t) dig_T1 << 1))) * ((int32_t) dig_T2)) >> 11;
	int32_t var2 = (((((adc_T >> 4) - ((int32_t) dig_T1)) * ((adc_T >> 4) - ((int32_t) dig_T1))) >> 12) * ((int32_t) dig_T3))
			>> 14;
	int32_t t_fine = var1 + var2;

	float T = (t_fine * 5 + 128) >> 8;
	return T / 100.0f;
}

float bme280_read_pressure(void) {
	uint8_t ctrl_meas = 0b00100101;
	i2c_mod2_write_register(bme280_address << 1, 0xF4, 1, &ctrl_meas, 1);
	HAL_Delay(10);

	uint8_t buf[6];
	i2c_mod2_read_register(bme280_address << 1, 0xF7, 1, buf, 6);

	int32_t adc_P = ((int32_t) buf[0] << 12) | ((int32_t) buf[1] << 4) | (buf[2] >> 4);
	int32_t adc_T = ((int32_t) buf[3] << 12) | ((int32_t) buf[4] << 4) | (buf[5] >> 4);

	int32_t var1 = ((((adc_T >> 3) - ((int32_t) dig_T1 << 1))) * ((int32_t) dig_T2)) >> 11;
	int32_t var2 = (((((adc_T >> 4) - ((int32_t) dig_T1)) * ((adc_T >> 4) - ((int32_t) dig_T1))) >> 12) * ((int32_t) dig_T3))
			>> 14;
	int32_t t_fine = var1 + var2;

	int64_t var1_p = ((int64_t) t_fine) - 128000;
	int64_t var2_p = var1_p * var1_p * (int64_t) dig_P6;
	var2_p = var2_p + ((var1_p * (int64_t) dig_P5) << 17);
	var2_p = var2_p + (((int64_t) dig_P4) << 35);
	var1_p = ((var1_p * var1_p * (int64_t) dig_P3) >> 8) + ((var1_p * (int64_t) dig_P2) << 12);
	var1_p = (((((int64_t) 1) << 47) + var1_p)) * ((int64_t) dig_P1) >> 33;

	if (var1_p == 0) {
		return 0;
	}

	int64_t p = 1048576 - adc_P;
	p = (((p << 31) - var2_p) * 3125) / var1_p;
	var1_p = (((int64_t) dig_P9) * (p >> 13) * (p >> 13)) >> 25;
	var2_p = (((int64_t) dig_P8) * p) >> 19;
	p = ((p + var1_p + var2_p) >> 8) + (((int64_t) dig_P7) << 4);

	return (float) p / 25600.0f;
}

float bme280_read_humidity(void) {
	uint8_t ctrl_hum = 0b00000001;
	i2c_mod2_write_register(bme280_address << 1, 0xF2, 1, &ctrl_hum, 1);

	uint8_t ctrl_meas = 0b00100101;
	i2c_mod2_write_register(bme280_address << 1, 0xF4, 1, &ctrl_meas, 1);
	HAL_Delay(10);

	uint8_t buf[8];
	i2c_mod2_read_register(bme280_address << 1, 0xF7, 1, buf, 8);

	int32_t adc_T = ((int32_t) buf[3] << 12) | ((int32_t) buf[4] << 4) | (buf[5] >> 4);
	int32_t adc_H = ((int32_t) buf[6] << 8) | buf[7];

	int32_t var1 = ((((adc_T >> 3) - ((int32_t) dig_T1 << 1))) * ((int32_t) dig_T2)) >> 11;
	int32_t var2 = (((((adc_T >> 4) - ((int32_t) dig_T1)) * ((adc_T >> 4) - ((int32_t) dig_T1))) >> 12) * ((int32_t) dig_T3))
			>> 14;
	int32_t t_fine = var1 + var2;

	int32_t v_x1_u32r = t_fine - ((int32_t) 76800);
	v_x1_u32r = (((((adc_H << 14) - (((int32_t) dig_H4) << 20) - (((int32_t) dig_H5) * v_x1_u32r)) + ((int32_t) 16384)) >> 15)
			* (((((((v_x1_u32r * ((int32_t) dig_H6)) >> 10) * (((v_x1_u32r * ((int32_t) dig_H3)) >> 11) + ((int32_t) 32768)))
					>> 10) + ((int32_t) 2097152)) * ((int32_t) dig_H2) + 8192) >> 14));

	v_x1_u32r = v_x1_u32r - (((((v_x1_u32r >> 15) * (v_x1_u32r >> 15)) >> 7) * ((int32_t) dig_H1)) >> 4);
	v_x1_u32r = (v_x1_u32r < 0 ? 0 : v_x1_u32r);
	v_x1_u32r = (v_x1_u32r > 419430400 ? 419430400 : v_x1_u32r);

	float h = (v_x1_u32r >> 12);
	return h / 1024.0f;
}

